"""Capsule (ERF/RIM) abstraction: LazyCapsule and in-memory load/save for archives."""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, cast

from pykotor.common.stream import BinaryReader
from pykotor.extract.file import FileResource, ResourceIdentifier, ResourceResult
from pykotor.resource.formats.erf import ERF, ERFType, write_erf
from pykotor.resource.formats.rim import RIM, write_rim
from pykotor.resource.type import ResourceType
from pykotor.tools.misc import is_any_erf_type_file, is_capsule_file, is_rim_file

if TYPE_CHECKING:
    import os

    from collections.abc import Iterator

    from typing_extensions import Self  # pyright: ignore[reportMissingModuleSource]


class LazyCapsule(FileResource):
    """LazyCapsule object is used for loading the list of resources stored in the .erf/.rim/.mod/.sav files used by the game.

    Resource data is not actually stored in memory by default but is instead loaded up on demand with the
    LazyCapsule.resource() method. Use the Capsule, RIM, or ERF classes if you want to solely work with capsules in memory.

    References:
    ----------
        See pykotor.resource.formats.erf.erf_data for engine addresses (K1 + TSL TODO). CExoEncapsulatedFile::CExoEncapsulatedFile, CExoKeyTable::AddEncapsulatedContents.


    """

    def __init__(
        self,
        path: os.PathLike | str,
        *,
        create_nonexisting: bool = False,
    ):
        """Initialize a Capsule object.

        Args:
        ----
            path: Path to the capsule file
            create_nonexisting: Whether to create the file if it doesn't exist

        Returns:
        -------
            self: The initialized Capsule object

        Processing Logic:
        ----------------
            - Check if the path points to a valid capsule file
            - If create_nonexisting is True and file doesn't exist:
                - Create RIM file if rim extension
                - Create ERF file if erf/mod extension
            - Initialize self._path and self._resources attributes
            - Reload resources from file.
        """
        ident = ResourceIdentifier.from_path(path)
        c_filepath = Path(path)
        if not is_capsule_file(c_filepath):
            msg = f"Invalid file extension in capsule filepath '{c_filepath}'."
            raise ValueError(msg)
        if create_nonexisting and not c_filepath.is_file():
            if is_rim_file(c_filepath):
                write_rim(RIM(), c_filepath)
            elif is_any_erf_type_file(c_filepath):
                write_erf(ERF(ERFType.from_extension(c_filepath.suffix)), c_filepath)
        super().__init__(*ident.unpack(), c_filepath.stat().st_size, 0x0, c_filepath)

    def __iter__(self) -> Iterator[FileResource]:
        yield from self.resources()

    def __len__(self):
        return len(self.resources())

    def resource(
        self,
        resref: str,
        restype: ResourceType,
    ) -> bytes | None:
        """Returns the bytes data of the specified resource. If the resource does not exist then returns None instead.

        Args:
        ----
            resref: The resource ResRef.
            restype: The resource type.

        Returns:
        -------
            bytes data of the resource or None if not existing.
        """
        query = ResourceIdentifier(resref, restype)
        resource: FileResource | None = next(
            (resource for resource in self.resources() if resource == query),
            None,
        )
        return resource.data() if resource else None

    def batch(
        self,
        queries: list[ResourceIdentifier],
    ) -> dict[ResourceIdentifier, ResourceResult | None]:
        """Batches queries against a capsule.

        Args:
        ----
            queries: list[ResourceIdentifier]: The queries to batch

        Returns:
        -------
            dict[ResourceIdentifier, ResourceResult | None]: The results for each query keyed by query

        Processing Logic:
        ----------------
            - Reloads capsule resources from erf/rim if reload is True
            - Initializes results dict to return
            - Opens capsule file as binary reader
            - Loops through queries
                - Sets result to None
                - Checks if resource exists in capsule
                - If so, seeks to offset, reads bytes and sets result
            - Closes reader
            - Returns results dict.
        """
        results: dict[ResourceIdentifier, ResourceResult | None] = {}
        with BinaryReader.from_file(self._filepath) as reader:
            for query in queries:
                results[query] = None

                resource: FileResource | None = next(
                    (resource for resource in self.resources() if resource == query),
                    None,
                )
                if resource is None:
                    continue

                reader.seek(resource.offset())
                data: bytes = reader.read_bytes(resource.size())
                results[query] = ResourceResult(
                    query.resname,
                    query.restype,
                    self._filepath,
                    data,
                )
        return results

    def contains(
        self,
        resref: str,
        restype: ResourceType,
    ) -> bool:
        """Check if a resource exists within this capsule.

        Args:
        ----
            resref: str: Resource reference
            restype: ResourceType: Resource type

        Returns:
        -------
            bool: True if resource exists, False otherwise

        Checks if a resource exists:
            - Constructs a ResourceIdentifier from resref and restype
            - Searches the ERF/RIM for a matching resource
            - Returns True if a match is found, False otherwise.
        """
        query = ResourceIdentifier(resref, restype)
        return bool(
            next(
                (resource for resource in self.resources() if resource == query),
                False,
            )
        )

    def info(
        self,
        resref: str,
        restype: ResourceType,
    ) -> FileResource | None:
        """Get file resource by reference and type.

        Args:
        ----
            resref: Resource reference as string
            restype: Resource type

        Returns:
        -------
            FileResource: Matched file resource

        Processing Logic:
        ----------------
            - Check if reload is True and call reload()
            - Create query object from resref and restype
            - Return first matching resource.
        """
        query = ResourceIdentifier(resref, restype)
        return next((resource for resource in self.resources() if resource == query), None)

    def resources(self) -> list[FileResource]:
        """Get the list of FileResources from the ERF/RIM file.

        Args:
        ----
            self: Capsule object to reload

        Returns:
        -------
            resources: list[FileResource] A list of FileResources from the ERF/RIM.

        Processing Logic:
        ----------------
            - Check if capsule exists on disk and print error if not
            - Check if file is empty (0 bytes) and return empty list if so
            - Open file and read header
            - Call appropriate reload method based on file type
            - Raise error if unknown file type.
        """
        # Check if file is empty (0 bytes) - empty files cannot be valid capsules
        if self._filepath.exists() and self._filepath.stat().st_size == 0:
            return []

        with BinaryReader.from_file(self._filepath) as reader:
            file_type = reader.read_string(4)
            reader.skip(4)  # file version

            if file_type in (member.value for member in ERFType):
                resources = self._load_erf(reader)
            elif file_type == "RIM ":
                resources = self._load_rim(reader)
            else:
                msg = f"File '{self._filepath}' must be a ERF/MOD/SAV/RIM capsule, '{self._filepath.suffix}' is not implemented."
                raise NotImplementedError(msg)
        # get_root_logger().debug("%s.resources() call, found %s total resources inside %s", self.__class__.__name__, len(resources), self._filepath)
        return resources

    def add(
        self,
        resname: str,
        restype: ResourceType,
        resdata: bytes,
    ):
        """Adds a resource to the capsule and writes the updated capsule to the disk.

        Args:
        ----
            resname: Name of the resource to add.
            restype: Type of the resource to add.
            resdata: Data of the resource to add.

        Returns:
        -------
            None: No value is returned.

        Processing Logic:
        ----------------
            - Checks if the file is RIM or a type of ERF
            - Reads the file as appropriate container
            - Calls set_data to add the resource
            - Writes the container back to the file.
        """

        def _add_to(container: RIM | ERF):
            container.set_data(resname, restype, resdata)
            for resource in self.resources():
                container.set_data(resource.resname(), resource.restype(), resource.data())

        if is_rim_file(self._filepath.name):
            container = RIM()
            _add_to(container)
            write_rim(container, self._filepath)
        elif is_any_erf_type_file(self._filepath.name):
            container = ERF(ERFType.from_extension(self._filepath))
            _add_to(container)
            write_erf(container, self._filepath)
        else:
            msg = f"File '{self._filepath}' is not a ERF/MOD/SAV/RIM capsule."
            raise NotImplementedError(msg)

    def delete(
        self,
        resname: str,
        restype: ResourceType,
    ):
        """Removes a resource from the capsule and writes the updated capsule to the disk.

        Args:
        ----
            resname: Name of the resource to add.
            restype: Type of the resource to add.

        Returns:
        -------
            None: No value is returned.

        Processing Logic:
        ----------------
            - Checks if the file is RIM or a type of ERF
            - Reads the file as appropriate container
            - Calls ERF.remove or RIM.remove
            - Writes the container back to the file.
        """

        def _remove_from(container: RIM | ERF):
            for resource in self.resources():
                if resource.resname().lower() == resname.lower() and resource.restype() is restype:
                    continue
                container.set_data(resource.resname(), resource.restype(), resource.data())

        if is_rim_file(self._filepath.name):
            container = RIM()
            _remove_from(container)
            write_rim(container, self._filepath)
        elif is_any_erf_type_file(self._filepath.name):
            container = ERF(ERFType.from_extension(self._filepath))
            _remove_from(container)
            write_erf(container, self._filepath)
        else:
            msg = f"File '{self._filepath}' is not a ERF/MOD/SAV/RIM capsule."
            raise NotImplementedError(msg)

    def as_cached_erf(self, erf_type: ERFType | None = None) -> ERF:
        erf: ERF = ERF() if erf_type is None else ERF(ERFType(erf_type))
        for resource in self.resources():
            erf.set_data(resource.resname(), resource.restype(), resource.data())
        return erf

    def as_cached_rim(self) -> RIM:
        rim: RIM = RIM()
        for resource in self.resources():
            rim.set_data(resource.resname(), resource.restype(), resource.data())
        return rim

    def as_cached(self) -> ERF | RIM:
        return self.as_cached_erf() if is_any_erf_type_file(self._filepath) else self.as_cached_rim()

    def _load_erf(
        self,
        reader: BinaryReader,
    ) -> list[FileResource]:
        """Loads an ERF resource file.

        Args:
        ----
            reader: BinaryReader - Reader for the ERF file

        Returns:
        -------
            resources: list[FileResource] A list of FileResources from the ERF.

        Processing Logic:
        ----------------
            - Skips header data
            - Reads entry count and offset tables
            - Loops through keys to read resource references, IDs and types
            - Seeks to resource data offset table
            - Loops to read offsets and sizes and populate resource objects.
        """

        resources: list[FileResource] = []
        reader.skip(8)
        entry_count = reader.read_uint32()
        reader.skip(4)
        offset_to_keys = reader.read_uint32()
        offset_to_resources = reader.read_uint32()

        resrefs: list[str] = []
        resids: list[int] = []
        restypes: list[ResourceType] = []
        reader.seek(offset_to_keys)

        for _ in range(entry_count):
            resref = reader.read_string(16).rstrip("\0")
            resrefs.append(resref)
            resids.append(reader.read_uint32())
            restype = reader.read_uint16()
            restypes.append(ResourceType.from_id(restype))
            reader.skip(2)

        reader.seek(offset_to_resources)
        for i in range(entry_count):
            res_offset: int = reader.read_uint32()
            res_size: int = reader.read_uint32()
            resources.append(FileResource(resrefs[i], restypes[i], res_size, res_offset, self._filepath))
        return resources

    def _load_rim(
        self,
        reader: BinaryReader,
    ) -> list[FileResource]:
        """Load resources from a rim file.

        Args:
        ----
            reader: BinaryReader: The binary reader to read data from

        Returns:
        -------
            resources: list[FileResource] A list of FileResources from the rim.

        Processing Logic:
        ----------------
            - Skip the first 4 bytes of unknown data
            - Read the entry count from the next 4 bytes
            - Read the offset to entries from the next 4 bytes
            - Seek to the offset to entries
            - Loop through each entry
                - Read the 16 byte resref string
                - Read the 4 byte resource type id and convert to enum
                - Skip the next 4 bytes of unknown data
                - Read the 4 byte offset
                - Read the 4 byte size
        """

        resources: list[FileResource] = []
        reader.skip(4)
        entry_count = reader.read_uint32()
        offset_to_entries = reader.read_uint32()

        reader.seek(offset_to_entries)

        for _ in range(entry_count):
            resref = reader.read_string(16).rstrip("\0")
            restype = ResourceType.from_id(reader.read_uint32())
            reader.skip(4)
            offset = reader.read_uint32()
            size = reader.read_uint32()
            resources.append(FileResource(resref, restype, size, offset, self._filepath))
        return resources


class Capsule(LazyCapsule):
    """Capsule object is used for loading the list of resources stored in the .erf/.rim/.mod/.sav files used by the game.

    Resource data is stored in memory on initialization and only reloaded when self.reload() is called or a `reload` argument is passed to relevant functions.
    """

    def __init__(
        self,
        path: os.PathLike | str,
        *,
        create_nonexisting: bool = False,
        reload: bool = True,
    ):
        """Initialize a Capsule object.

        Args:
        ----
            path: Path to the capsule file
            create_nonexisting: Whether to create the file if it doesn't exist

        Returns:
        -------
            self: The initialized Capsule object

        Processing Logic:
        ----------------
            - Check if the path points to a valid capsule file
            - If create_nonexisting is True and file doesn't exist:
                - Create RIM file if rim extension
                - Create ERF file if erf/mod extension
            - Initialize self._path and self._resources attributes
            - Reload resources from file.
        """
        self._resources: list[FileResource] = []
        super().__init__(path, create_nonexisting=create_nonexisting)
        if reload:
            self.reload()

    def resources(
        self,
        *,
        reload: bool = False,
    ) -> list[FileResource]:
        if reload:
            self.reload()
        return self._resources

    def resource(
        self,
        resref: str,
        restype: ResourceType,
        *,
        reload: bool = False,
    ) -> bytes | None:
        """Returns the bytes data of the specified resource. If the resource does not exist then returns None instead.

        Args:
        ----
            resref: The resource ResRef.
            restype: The resource type.
            reload: If True Capsule will reload the ERF before opening rather than using cached offsets.

        Returns:
        -------
            bytes data of the resource.
        """
        if reload:
            self.reload()
        return super().resource(resref, restype)

    def batch(
        self,
        queries: list[ResourceIdentifier],
        *,
        reload: bool = False,
    ) -> dict[ResourceIdentifier, ResourceResult | None]:
        """Batches queries against a capsule.

        Args:
        ----
            queries: list[ResourceIdentifier]: The queries to batch
            reload: bool = False: Whether to reload the capsule resources

        Returns:
        -------
            dict[ResourceIdentifier, ResourceResult | None]: The results for each query keyed by query

        Processing Logic:
        ----------------
            - Reloads capsule resources from erf/rim if reload is True
            - Initializes results dict to return
            - Opens capsule file as binary reader
            - Loops through queries
                - Sets result to None
                - Checks if resource exists in capsule
                - If so, seeks to offset, reads bytes and sets result
            - Closes reader
            - Returns results dict.
        """
        if reload:
            self.reload()
        return super().batch(queries)

    def contains(
        self,
        resref: str,
        restype: ResourceType,
        *,
        reload: bool = False,
    ) -> bool:
        """Check if a resource exists.

        Args:
        ----
            resref: str: Resource reference
            restype: ResourceType: Resource type
            reload: bool: Reload resources cache

        Returns:
        -------
            bool: True if resource exists, False otherwise

        Checks if a resource exists:
            - Constructs a ResourceIdentifier from resref and restype
            - Searches self._resources for a matching resource
            - Returns True if a match is found, False otherwise.
        """
        if reload:
            self.reload()
        return super().contains(resref, restype)

    def info(
        self,
        resref: str,
        restype: ResourceType,
        *,
        reload: bool = False,
    ) -> FileResource | None:
        """Get file resource by reference and type.

        Args:
        ----
            resref: Resource reference as string
            restype: Resource type
            reload: Reload resources if True (kwarg)

        Returns:
        -------
            FileResource: Matched file resource

        Processing Logic:
        ----------------
            - Check if reload is True and call reload()
            - Create query object from resref and restype
            - Return first matching resource from internal list.
        """
        if reload:
            self.reload()
        return super().info(resref, restype)

    def reload(self):
        """Reload the list of resource info linked from the module file.

        Args:
        ----
            self: Capsule object to reload

        Returns:
        -------
            None: Reloading is done in-place

        Processing Logic:
        ----------------
            - Check if capsule exists on disk and print error if not
            - Open file and read header
            - Call appropriate reload method based on file type
            - Raise error if unknown file type.
        """
        self._internal = True
        self._resources = super().resources()
        self._internal = False

    def delete(
        self,
        resname: str,
        restype: ResourceType,
    ):
        result = super().delete(resname, restype)
        self.reload()
        return result

    def add(
        self,
        resname: str,
        restype: ResourceType,
        resdata: bytes,
    ):
        """Adds a resource to the capsule and writes the updated capsule to the disk.

        Args:
        ----
            resname: Name of the resource to add in one line.
            restype: Type of the resource to add in one line.
            resdata: Data of the resource to add in one line.

        Returns:
        -------
            None: No value is returned in one line.

        Processing Logic:
        ----------------
            - Checks if the file is RIM or a type of ERF
            - Reads the file as appropriate container
            - Calls set_data to add the resource
            - Writes the container back to the file.
        """
        result = super().add(resname, restype, resdata)
        self.reload()
        return result

    @classmethod
    def from_data(cls, data: bytes, resname: str, restype: ResourceType) -> Self:
        """Initialize a Capsule object from raw data bytes.

        Args:
        ----
            data: The raw bytes of the capsule file.
            resname: The resource name for identification.
            restype: The resource type.

        Returns:
        -------
            Capsule: The initialized Capsule object.
        """
        with BinaryReader.from_bytes(data) as reader:
            capsule = cast("cls", object())
            capsule.__class__ = cls
            file_type = reader.read_string(4)
            reader.skip(4)
            if file_type in (member.value for member in ERFType):
                capsule._resources = capsule._load_erf(reader)  # noqa: SLF001
            elif file_type == "RIM ":
                capsule._resources = capsule._load_rim(reader)  # noqa: SLF001
            else:
                msg = f"Data provided is not a valid ERF/MOD/SAV/RIM format, found '{file_type}'."
                raise ValueError(msg)
            return capsule



"""NSS (NWScript) lexer: PLY lex tokenization and keyword/operator mapping."""

from __future__ import annotations
from pykotor.resource.formats._base import BiowareResource

from typing import ClassVar

from ply import lex

from pykotor.common.script import DataType
from pykotor.resource.formats.ncs import NCSInstructionType
from pykotor.resource.formats.ncs.compiler.classes import (
    BinaryOperatorMapping,
    ControlKeyword,
    FloatExpression,
    Identifier,
    IntExpression,
    ObjectExpression,
    OperatorMapping,
    StringExpression,
    UnaryOperatorMapping,
)


class NssLexer(BiowareResource):
    """NSS (NWScript Source) lexer/tokenizer.

    Tokenizes NSS source code into tokens for parsing. Handles keywords, operators,
    literals, identifiers, and special values (OBJECTSELF, OBJECTINVALID, etc.).

    References:
    ----------

    Derivations and Other Implementations:
    -------------------------------------
        https://github.com/th3w1zard1/KotOR.js/tree/master/src/nwscript/NWScriptCompiler.ts (Token handling)
        PLY (Python Lex-Yacc) library for lexer generation

    """

    def __init__(
        self,
        errorlog=lex.NullLogger(),  # noqa: B008
        *,
        nowarn=True,
    ):
        self.lexer: lex.Lexer = lex.lex(module=self, errorlog=errorlog, nowarn=nowarn)
        self.lexer.begin("INITIAL")

    tokens: ClassVar[list[str]] = [
        "STRING_VALUE",
        "INT_VALUE",
        "FLOAT_VALUE",
        "IDENTIFIER",
        "INT_TYPE",
        "INT_HEX_VALUE",
        "FLOAT_TYPE",
        "OBJECT_TYPE",
        "VOID_TYPE",
        "EVENT_TYPE",
        "EFFECT_TYPE",
        "ITEMPROPERTY_TYPE",
        "LOCATION_TYPE",
        "STRING_TYPE",
        "TALENT_TYPE",
        "VECTOR_TYPE",
        "ACTION_TYPE",
        "BREAK_CONTROL",
        "CASE_CONTROL",
        "DEFAULT_CONTROL",
        "DO_CONTROL",
        "ELSE_CONTROL",
        "SWITCH_CONTROL",
        "WHILE_CONTROL",
        "FOR_CONTROL",
        "IF_CONTROL",
        "TRUE_VALUE",
        "FALSE_VALUE",
        "OBJECTSELF_VALUE",
        "OBJECTINVALID_VALUE",
        "ADD",
        "MINUS",
        "MULTIPLY",
        "DIVIDE",
        "MOD",
        "EQUALS",
        "NOT_EQUALS",
        "GREATER_THAN",
        "LESS_THAN",
        "LESS_THAN_OR_EQUALS",
        "GREATER_THAN_OR_EQUALS",
        "AND",
        "OR",
        "NOT",
        "BITWISE_AND",
        "BITWISE_OR",
        "BITWISE_LEFT",
        "BITWISE_RIGHT",
        "BITWISE_UNSIGNED_RIGHT",
        "BITWISE_XOR",
        "BITWISE_NOT",
        "INCLUDE",
        "RETURN",
        "ADDITION_ASSIGNMENT_OPERATOR",
        "SUBTRACTION_ASSIGNMENT_OPERATOR",
        "MULTIPLICATION_ASSIGNMENT_OPERATOR",
        "DIVISION_ASSIGNMENT_OPERATOR",
        "MOD_ASSIGNMENT_OPERATOR",
        "BITWISE_AND_ASSIGNMENT_OPERATOR",
        "BITWISE_OR_ASSIGNMENT_OPERATOR",
        "BITWISE_XOR_ASSIGNMENT_OPERATOR",
        "BITWISE_LEFT_ASSIGNMENT_OPERATOR",
        "BITWISE_RIGHT_ASSIGNMENT_OPERATOR",
        "BITWISE_UNSIGNED_RIGHT_ASSIGNMENT_OPERATOR",
        "CONTINUE_CONTROL",
        "STRUCT",
        "INCREMENT",
        "DECREMENT",
        "NOP",
        "CONST",
    ]

    literals: ClassVar[list[str]] = [
        "{",
        "}",
        "(",
        ")",
        ";",
        "=",
        ",",
        ":",
        ".",
        "[",
        "]",
        "?",
    ]

    t_ignore: str = " \t\r"

    def t_NEWLINE(self, t):
        r"\n+"  # noqa: D300, D400, D415
        t.lexer.lineno += len(t.value)

    def t_NOP(self, t):
        r"[Nn][Oo][Pp]"  # noqa: D300, D400, D415, D403
        return t

    def t_COMMENT(self, t):
        r"//[^\n]*\n"  # noqa: D300, D400, D415
        t.lexer.lineno += 1

    def t_MULTILINE_COMMENT(self, t):
        r"\/\*(\*(?!\/)|[^*])*\*\/"  # noqa: D300, D400, D415
        t.lexer.lineno += t.value.count("\n")

    def t_INCLUDE(self, t):
        r"\#include"  # noqa: D300, D400, D415
        return t

    def t_OBJECTSELF_VALUE(self, t):
        r"OBJECT_SELF\b"  # noqa: D300, D400, D415
        t.value = ObjectExpression(0)
        return t

    def t_OBJECTINVALID_VALUE(self, t):
        r"OBJECT_INVALID\b"  # noqa: D300, D400, D415
        t.value = ObjectExpression(1)
        return t

    def t_TRUE_VALUE(self, t):
        r"TRUE\b"  # noqa: D300, D400, D415
        t.value = IntExpression(1)
        return t

    def t_FALSE_VALUE(self, t):
        r"FALSE\b"  # noqa: D300, D400, D415
        t.value = IntExpression(0)
        return t

    # region Control Tokens
    def t_BREAK_CONTROL(self, t):
        r"break\b"  # noqa: D300, D400, D415
        t.value = ControlKeyword.BREAK
        return t

    def t_CONTINUE_CONTROL(self, t):
        r"continue\b"  # noqa: D300, D400, D415
        return t

    def t_CASE_CONTROL(self, t):
        r"case\b"  # noqa: D300, D400, D415
        t.value = ControlKeyword.CASE
        return t

    def t_DEFAULT_CONTROL(self, t):
        r"default\b"  # noqa: D300, D400, D415
        t.value = ControlKeyword.DEFAULT
        return t

    def t_DO_CONTROL(self, t):
        r"do\b"  # noqa: D300, D400, D415
        t.value = ControlKeyword.DO
        return t

    def t_ELSE_CONTROL(self, t):
        r"else\b"  # noqa: D300, D400, D415
        t.value = ControlKeyword.ELSE
        return t

    def t_SWITCH_CONTROL(self, t):
        r"switch\b"  # noqa: D300, D400, D415
        t.value = ControlKeyword.SWITCH
        return t

    def t_WHILE_CONTROL(self, t):
        r"while\b"  # noqa: D300, D400, D415
        t.value = ControlKeyword.WHILE
        return t

    def t_FOR_CONTROL(self, t):
        r"for\b"  # noqa: D300, D400, D415
        t.value = ControlKeyword.FOR
        return t

    def t_IF_CONTROL(self, t):
        r"if\b"  # noqa: D300, D400, D415
        t.value = ControlKeyword.IF
        return t

    def t_RETURN(self, t):
        r"return\b"  # noqa: D300, D400, D415
        t.value = ControlKeyword.RETURN
        return t

    def t_CONST(self, t):
        r"const\b"  # noqa: D300, D400, D415
        return t

    # endregion

    # region Type Tokens
    def t_STRUCT(self, t):
        r"struct\b"  # noqa: D300, D400, D415
        t.value = DataType.STRUCT
        return t

    def t_INT_TYPE(self, t):
        r"int\b"  # noqa: D300, D400, D415
        t.value = DataType.INT
        return t

    def t_FLOAT_TYPE(self, t):
        r"float\b"  # noqa: D300, D400, D415
        t.value = DataType.FLOAT
        return t

    def t_OBJECT_TYPE(self, t):
        r"object\b"  # noqa: D300, D400, D415
        t.value = DataType.OBJECT
        return t

    def t_VOID_TYPE(self, t):
        r"void\b"  # noqa: D300, D400, D415
        t.value = DataType.VOID
        return t

    def t_EVENT_TYPE(self, t):
        r"event\b"  # noqa: D300, D400, D415
        t.value = DataType.EVENT
        return t

    def t_EFFECT_TYPE(self, t):
        r"effect\b"  # noqa: D300, D400, D415
        t.value = DataType.EFFECT
        return t

    def t_ITEMPROPERTY_TYPE(self, t):
        r"itemproperty\b"  # noqa: D300, D400, D415
        t.value = DataType.ITEMPROPERTY
        return t

    def t_LOCATION_TYPE(self, t):
        r"location\b"  # noqa: D300, D400, D415
        t.value = DataType.LOCATION
        return t

    def t_STRING_TYPE(self, t):
        r"string\b"  # noqa: D300, D400, D415
        t.value = DataType.STRING
        return t

    def t_TALENT_TYPE(self, t):
        r"talent\b"  # noqa: D300, D400, D415
        t.value = DataType.TALENT
        return t

    def t_ACTION_TYPE(self, t):
        r"action\b"  # noqa: D300, D400, D415
        t.value = DataType.ACTION
        return t

    def t_VECTOR_TYPE(self, t):
        r"vector\b"  # noqa: D300, D400, D415
        t.value = DataType.VECTOR
        return t

    # endregion

    # region Value Tokens
    def t_IDENTIFIER(self, t):
        "[a-zA-Z_]+[a-zA-Z0-9_]*"  # noqa: D300, D400, D415
        t.value = Identifier(t.value)
        return t

    def t_STRING_VALUE(self, t):
        r"\"([^\"\\]|\\.)*\""  # noqa: D300, D400, D415, D210
        # Handle escape sequences: \\, \", \n, \r, \t
        value = t.value[1:-1]  # Remove quotes
        # Basic escape sequence handling (order matters - do \\ first)
        value = value.replace("\\\\", "\\")
        value = value.replace('\\"', '"')
        value = value.replace("\\n", "\n")
        value = value.replace("\\r", "\r")
        value = value.replace("\\t", "\t")
        t.value = StringExpression(value)
        return t

    def t_FLOAT_VALUE(self, t):
        r"[0-9]+\.[0-9]+f?|[0-9]f"  # noqa: D300, D400, D415
        t.value = FloatExpression(float(t.value.replace("f", "")))
        return t

    def t_INT_HEX_VALUE(self, t):
        "0x[0-9a-fA-F]+"  # noqa: D300, D400, D415
        t.value = IntExpression(int(t.value, 16))
        return t

    def t_INT_VALUE(self, t):
        "[0-9]+"  # noqa: D300, D400, D415
        t.value = IntExpression(int(t.value))
        return t

    # endregion

    def t_INCREMENT(self, t):
        r"\+\+"  # noqa: D300, D400, D415
        t.value = OperatorMapping(
            unary=[
                UnaryOperatorMapping(NCSInstructionType.INCxSP, DataType.INT),
            ],
            binary=[],
        )
        return t

    def t_DECREMENT(self, t):
        r"\-\-"  # noqa: D300, D400, D415
        t.value = OperatorMapping(
            unary=[
                UnaryOperatorMapping(NCSInstructionType.DECxSP, DataType.INT),
            ],
            binary=[],
        )
        return t

    def t_ADDITION_ASSIGNMENT_OPERATOR(self, t):
        r"\+\="  # noqa: D300, D400, D415
        return t

    def t_SUBTRACTION_ASSIGNMENT_OPERATOR(self, t):
        r"\-\="  # noqa: D300, D400, D415
        return t

    def t_MULTIPLICATION_ASSIGNMENT_OPERATOR(self, t):
        r"\*\="  # noqa: D300, D400, D415
        return t

    def t_DIVISION_ASSIGNMENT_OPERATOR(self, t):
        r"/\="  # noqa: D300, D400, D415
        return t

    def t_MOD_ASSIGNMENT_OPERATOR(self, t):
        r"%\="  # noqa: D300, D400, D415
        return t

    def t_BITWISE_AND_ASSIGNMENT_OPERATOR(self, t):
        r"&\="  # noqa: D300, D400, D415
        return t

    def t_BITWISE_OR_ASSIGNMENT_OPERATOR(self, t):
        r"\|\="  # noqa: D300, D400, D415
        return t

    def t_BITWISE_XOR_ASSIGNMENT_OPERATOR(self, t):
        r"\^\="  # noqa: D300, D400, D415
        return t

    def t_BITWISE_LEFT_ASSIGNMENT_OPERATOR(self, t):
        r"<<\="  # noqa: D300, D400, D415
        return t

    def t_BITWISE_RIGHT_ASSIGNMENT_OPERATOR(self, t):
        r">>\="  # noqa: D300, D400, D415
        return t

    def t_BITWISE_UNSIGNED_RIGHT_ASSIGNMENT_OPERATOR(self, t):
        r">>>\="  # noqa: D300, D400, D415
        return t

    # region Operators
    def t_BITWISE_LEFT(self, t):
        "<<"  # noqa: D300, D400, D415
        t.value = OperatorMapping(
            unary=[],
            binary=[
                BinaryOperatorMapping(NCSInstructionType.SHLEFTII, DataType.INT, DataType.INT, DataType.INT),
            ],
        )
        return t

    def t_BITWISE_UNSIGNED_RIGHT(self, t):
        ">>>"  # noqa: D300, D400, D415
        t.value = OperatorMapping(
            unary=[],
            binary=[
                BinaryOperatorMapping(NCSInstructionType.USHRIGHTII, DataType.INT, DataType.INT, DataType.INT),
            ],
        )
        return t

    def t_BITWISE_RIGHT(self, t):
        ">>"  # noqa: D300, D400, D415
        t.value = OperatorMapping(
            unary=[],
            binary=[
                BinaryOperatorMapping(NCSInstructionType.SHRIGHTII, DataType.INT, DataType.INT, DataType.INT),
            ],
        )
        return t

    def t_ADD(self, t):
        r"\+"  # noqa: D300, D400, D415
        t.value = OperatorMapping(
            unary=[],
            binary=[
                BinaryOperatorMapping(NCSInstructionType.ADDII, DataType.INT, DataType.INT, DataType.INT),
                BinaryOperatorMapping(NCSInstructionType.ADDIF, DataType.INT, DataType.INT, DataType.FLOAT),
                BinaryOperatorMapping(NCSInstructionType.ADDFI, DataType.FLOAT, DataType.FLOAT, DataType.INT),
                BinaryOperatorMapping(NCSInstructionType.ADDFF, DataType.FLOAT, DataType.FLOAT, DataType.FLOAT),
                BinaryOperatorMapping(NCSInstructionType.ADDVV, DataType.VECTOR, DataType.VECTOR, DataType.VECTOR),
                BinaryOperatorMapping(NCSInstructionType.ADDSS, DataType.STRING, DataType.STRING, DataType.STRING),
            ],
        )
        return t

    def t_MINUS(self, t):
        "-"  # noqa: D300, D415, D400
        t.value = OperatorMapping(
            unary=[
                UnaryOperatorMapping(NCSInstructionType.NEGI, DataType.INT),
                UnaryOperatorMapping(NCSInstructionType.NEGF, DataType.FLOAT),
            ],
            binary=[
                BinaryOperatorMapping(NCSInstructionType.SUBII, DataType.INT, DataType.INT, DataType.INT),
                BinaryOperatorMapping(NCSInstructionType.SUBIF, DataType.INT, DataType.INT, DataType.FLOAT),
                BinaryOperatorMapping(NCSInstructionType.SUBFI, DataType.FLOAT, DataType.FLOAT, DataType.INT),
                BinaryOperatorMapping(NCSInstructionType.SUBFF, DataType.FLOAT, DataType.FLOAT, DataType.FLOAT),
                BinaryOperatorMapping(NCSInstructionType.SUBVV, DataType.VECTOR, DataType.VECTOR, DataType.VECTOR),
            ],
        )
        return t

    def t_MULTIPLY(self, t):
        r"\*"  # noqa: D300, D400, D415
        t.value = OperatorMapping(
            unary=[],
            binary=[
                BinaryOperatorMapping(NCSInstructionType.MULII, DataType.INT, DataType.INT, DataType.INT),
                BinaryOperatorMapping(NCSInstructionType.MULIF, DataType.INT, DataType.INT, DataType.FLOAT),
                BinaryOperatorMapping(NCSInstructionType.MULFI, DataType.FLOAT, DataType.FLOAT, DataType.INT),
                BinaryOperatorMapping(NCSInstructionType.MULFF, DataType.FLOAT, DataType.FLOAT, DataType.FLOAT),
                BinaryOperatorMapping(NCSInstructionType.MULVF, DataType.VECTOR, DataType.VECTOR, DataType.FLOAT),
                BinaryOperatorMapping(NCSInstructionType.MULFV, DataType.VECTOR, DataType.FLOAT, DataType.VECTOR),
            ],
        )
        return t

    def t_DIVIDE(self, t):
        "/"  # noqa: D300, D400, D415
        t.value = OperatorMapping(
            unary=[],
            binary=[
                BinaryOperatorMapping(NCSInstructionType.DIVII, DataType.INT, DataType.INT, DataType.INT),
                BinaryOperatorMapping(NCSInstructionType.DIVIF, DataType.INT, DataType.INT, DataType.FLOAT),
                BinaryOperatorMapping(NCSInstructionType.DIVFI, DataType.FLOAT, DataType.FLOAT, DataType.INT),
                BinaryOperatorMapping(NCSInstructionType.DIVFF, DataType.FLOAT, DataType.FLOAT, DataType.FLOAT),
                BinaryOperatorMapping(NCSInstructionType.DIVVF, DataType.VECTOR, DataType.VECTOR, DataType.FLOAT),
                BinaryOperatorMapping(NCSInstructionType.DIVFV, DataType.VECTOR, DataType.FLOAT, DataType.VECTOR),
            ],
        )
        return t

    def t_MOD(self, t):
        r"\%"  # noqa: D300, D400, D415
        t.value = OperatorMapping(
            unary=[],
            binary=[
                BinaryOperatorMapping(NCSInstructionType.MODII, DataType.INT, DataType.INT, DataType.INT),
            ],
        )
        return t

    def t_EQUALS(self, t):
        r"\=\="  # noqa: D300, D400, D415
        t.value = OperatorMapping(
            unary=[],
            binary=[
                BinaryOperatorMapping(NCSInstructionType.EQUALII, DataType.INT, DataType.INT, DataType.INT),
                BinaryOperatorMapping(NCSInstructionType.EQUALFF, DataType.INT, DataType.FLOAT, DataType.FLOAT),
                BinaryOperatorMapping(NCSInstructionType.EQUALOO, DataType.INT, DataType.OBJECT, DataType.OBJECT),
                BinaryOperatorMapping(NCSInstructionType.EQUALSS, DataType.INT, DataType.STRING, DataType.STRING),
            ],
        )
        return t

    def t_NOT_EQUALS(self, t):
        r"\!="  # noqa: D300, D400, D415
        t.value = OperatorMapping(
            unary=[],
            binary=[
                BinaryOperatorMapping(NCSInstructionType.NEQUALII, DataType.INT, DataType.INT, DataType.INT),
                BinaryOperatorMapping(NCSInstructionType.NEQUALFF, DataType.INT, DataType.FLOAT, DataType.FLOAT),
                BinaryOperatorMapping(NCSInstructionType.NEQUALOO, DataType.INT, DataType.OBJECT, DataType.OBJECT),
                BinaryOperatorMapping(NCSInstructionType.NEQUALSS, DataType.INT, DataType.STRING, DataType.STRING),
            ],
        )
        return t

    def t_GREATER_THAN_OR_EQUALS(self, t):
        r">\="  # noqa: D300, D400, D415
        t.value = OperatorMapping(
            unary=[],
            binary=[
                BinaryOperatorMapping(NCSInstructionType.GEQII, DataType.INT, DataType.INT, DataType.INT),
                BinaryOperatorMapping(NCSInstructionType.GEQFF, DataType.INT, DataType.FLOAT, DataType.FLOAT),
            ],
        )
        return t

    def t_GREATER_THAN(self, t):
        ">"  # noqa: D300, D400, D415
        t.value = OperatorMapping(
            unary=[],
            binary=[
                BinaryOperatorMapping(NCSInstructionType.GTII, DataType.INT, DataType.INT, DataType.INT),
                BinaryOperatorMapping(NCSInstructionType.GTFF, DataType.INT, DataType.FLOAT, DataType.FLOAT),
            ],
        )
        return t

    def t_LESS_THAN_OR_EQUALS(self, t):
        r"\<="  # noqa: D300, D400, D415
        t.value = OperatorMapping(
            unary=[],
            binary=[
                BinaryOperatorMapping(NCSInstructionType.LEQII, DataType.INT, DataType.INT, DataType.INT),
                BinaryOperatorMapping(NCSInstructionType.LEQFF, DataType.INT, DataType.FLOAT, DataType.FLOAT),
            ],
        )
        return t

    def t_LESS_THAN(self, t):
        r"\<"  # noqa: D300, D400, D415
        t.value = OperatorMapping(
            unary=[],
            binary=[
                BinaryOperatorMapping(NCSInstructionType.LTII, DataType.INT, DataType.INT, DataType.INT),
                BinaryOperatorMapping(NCSInstructionType.LTFF, DataType.INT, DataType.FLOAT, DataType.FLOAT),
            ],
        )
        return t

    def t_AND(self, t):
        "&&"  # noqa: D300, D400, D415
        t.value = OperatorMapping(
            unary=[],
            binary=[
                BinaryOperatorMapping(NCSInstructionType.LOGANDII, DataType.INT, DataType.INT, DataType.INT),
            ],
        )
        return t

    def t_OR(self, t):
        r"\|\|"  # noqa: D300, D400, D415
        t.value = OperatorMapping(
            unary=[],
            binary=[
                BinaryOperatorMapping(NCSInstructionType.LOGORII, DataType.INT, DataType.INT, DataType.INT),
            ],
        )
        return t

    def t_NOT(self, t):
        r"\!"  # noqa: D300, D400
        t.value = OperatorMapping(
            unary=[
                UnaryOperatorMapping(NCSInstructionType.NOTI, DataType.INT),
            ],
            binary=[],
        )
        return t

    def t_BITWISE_AND(self, t):
        "&"  # noqa: D300, D400, D415
        t.value = OperatorMapping(
            unary=[],
            binary=[
                BinaryOperatorMapping(NCSInstructionType.BOOLANDII, DataType.INT, DataType.INT, DataType.INT),
            ],
        )
        return t

    def t_BITWISE_OR(self, t):
        r"\|"  # noqa: D300, D400, D415
        t.value = OperatorMapping(
            unary=[],
            binary=[
                BinaryOperatorMapping(NCSInstructionType.INCORII, DataType.INT, DataType.INT, DataType.INT),
            ],
        )
        return t

    def t_BITWISE_XOR(self, t):
        r"\^"  # noqa: D300, D400, D415
        t.value = OperatorMapping(
            unary=[],
            binary=[
                BinaryOperatorMapping(NCSInstructionType.EXCORII, DataType.INT, DataType.INT, DataType.INT),
            ],
        )
        return t

    def t_BITWISE_NOT(self, t):
        r"\~"  # noqa: D300, D400, D415
        t.value = OperatorMapping(
            unary=[
                UnaryOperatorMapping(NCSInstructionType.COMPI, DataType.INT),
            ],
            binary=[],
        )
        return t

    # endregion



"""NSS (NWScript) Language Server.

A language server for NSS scripts that provides diagnostics, completions,
hover information, and document symbols. Designed to run in a subprocess
to avoid blocking the UI thread and to eliminate GIL contention.

PERFORMANCE CRITICAL:
The PLY YACC parser generation takes ~900ms on first initialization.
This module caches the parser instance so this cost is paid only once
per subprocess lifetime, not on every keystroke.

Usage:
    # Start language server in subprocess
    from multiprocessing import Process, Queue
    request_queue = Queue()
    response_queue = Queue()
    server_process = Process(
        target=NSSLanguageServer.run_server,
        args=(request_queue, response_queue, is_tsl)
    )
    server_process.start()

    # Send requests
    request_queue.put({
        'id': 1,
        'method': 'analyze',
        'params': {'text': '...', 'filepath': '...'}
    })

    # Get responses
    response = response_queue.get(timeout=5.0)

References:
----------
        Based on swkotor.exe NWScript structure:
        - NCS (NWScript Compiled Script) bytecode format
        - Language Server Protocol specification

        Note: This is a language server for NSS (NWScript Source) files, providing
        diagnostics, completions, and hover information. NSS compiles to NCS bytecode
        which is executed by the game engine.

"""

from __future__ import annotations
from pykotor.resource.formats._base import BiowareResource

import re
import traceback

from dataclasses import dataclass, field
from enum import IntEnum
from pathlib import Path
from typing import TYPE_CHECKING, Any

from ply import yacc

from pykotor.resource.formats.ncs.compiler.classes import (
    CompileError,
    FunctionDefinition,
    FunctionForwardDeclaration,
    GlobalVariableDeclaration,
    GlobalVariableInitialization,
    StructDefinition,
)
from pykotor.resource.formats.ncs.compiler.lexer import NssLexer
from pykotor.resource.formats.ncs.compiler.parser import NssParser

if TYPE_CHECKING:
    from multiprocessing import Queue

    from pykotor.common.script import ScriptConstant, ScriptFunction
    from pykotor.resource.formats.ncs.compiler.classes import CodeRoot


class DiagnosticSeverity(IntEnum):
    """Severity levels for diagnostics (matches LSP specification)."""

    ERROR = 1
    WARNING = 2
    INFORMATION = 3
    HINT = 4


@dataclass
class Position(BiowareResource):
    """A position in a text document (0-indexed)."""

    line: int
    character: int


@dataclass
class Range(BiowareResource):
    """A range in a text document."""

    start: Position
    end: Position


@dataclass
class Diagnostic(BiowareResource):
    """A diagnostic (error, warning, etc.) in a document."""

    range: Range
    message: str
    severity: DiagnosticSeverity = DiagnosticSeverity.ERROR
    code: str | None = None
    source: str = "nss"
    suggestions: list[str] = field(default_factory=list)


@dataclass
class DocumentSymbol(BiowareResource):
    """A symbol in a document (function, struct, variable, etc.)."""

    name: str
    kind: str  # 'function', 'struct', 'variable', 'parameter'
    range: Range
    selection_range: Range
    detail: str = ""
    children: list[DocumentSymbol] = field(default_factory=list)


@dataclass
class CompletionItem(BiowareResource):
    """An auto-completion suggestion."""

    label: str
    kind: str  # 'function', 'constant', 'keyword', 'variable'
    detail: str = ""
    documentation: str = ""
    insert_text: str = ""
    sort_text: str = ""


@dataclass
class HoverInfo(BiowareResource):
    """Hover information for a symbol."""

    contents: str  # Markdown-formatted content
    range: Range | None = None


@dataclass
class AnalysisResult(BiowareResource):
    """Complete analysis result for a document."""

    diagnostics: list[Diagnostic] = field(default_factory=list)
    symbols: list[DocumentSymbol] = field(default_factory=list)
    parse_successful: bool = False
    ast: CodeRoot | None = None


class NSSLanguageServer(BiowareResource):
    """Language server for NSS scripts.

    Provides diagnostics, completions, hover, and document symbols.
    Designed to be instantiated once per subprocess and reused for
    multiple analysis requests.

    Attributes:
        functions: List of built-in script functions
        constants: List of built-in script constants
        library: Dictionary of script library contents
        is_tsl: Whether this is for TSL (True) or K1 (False)
    """

    # Singleton parser and lexer - created once, reused forever
    _parser: NssParser | None = None
    _lexer: NssLexer | None = None
    _initialized_for_tsl: bool | None = None

    def __init__(
        self,
        functions: list[ScriptFunction] | None = None,
        constants: list[ScriptConstant] | None = None,
        library: dict[str, bytes] | None = None,
        is_tsl: bool = False,
    ):
        """Initialize the language server.

        Args:
            functions: Built-in script functions
            constants: Built-in script constants
            library: Script library for includes
            is_tsl: Whether this is for TSL (K2) or K1
        """
        self.functions = functions or []
        self.constants = constants or []
        self.library = library or {}
        self.is_tsl = is_tsl
        self._current_filepath: Path | None = None

        # Initialize parser on first use (expensive, ~900ms)
        self._ensure_parser_initialized()

    def _ensure_parser_initialized(self):
        """Ensure parser is initialized (only once per process)."""
        if NSSLanguageServer._parser is None or NSSLanguageServer._initialized_for_tsl != self.is_tsl:
            # Create parser with error suppression
            NSSLanguageServer._parser = NssParser(
                functions=self.functions,
                constants=self.constants,
                library=self.library,
                library_lookup=None,
                errorlog=yacc.NullLogger(),
            )
            NSSLanguageServer._lexer = NssLexer()
            NSSLanguageServer._initialized_for_tsl = self.is_tsl

    def _get_parser(self) -> NssParser:
        """Get the cached parser, updating its functions/constants if needed."""
        assert NSSLanguageServer._parser is not None
        parser = NSSLanguageServer._parser

        # Update parser references (cheap operation, no table regeneration)
        parser.functions = self.functions
        parser.constants = self.constants
        parser.library = self.library
        if self._current_filepath is not None:
            parser.library_lookup = [self._current_filepath.parent]
        else:
            parser.library_lookup = []

        return parser

    def _create_lexer(self) -> NssLexer:
        """Create a fresh lexer for each parse (lexer state needs reset)."""
        return NssLexer()

    def analyze(
        self,
        text: str,
        filepath: str | Path | None = None,
    ) -> AnalysisResult:
        """Analyze NSS source code and return diagnostics and symbols.

        This is the main entry point for document analysis. It parses the
        source code and extracts diagnostics (errors/warnings) and document
        symbols (functions, structs, variables).

        Args:
            text: NSS source code to analyze
            filepath: Optional file path for include resolution

        Returns:
            AnalysisResult with diagnostics and symbols
        """
        self._current_filepath = Path(filepath) if filepath else None
        result = AnalysisResult()

        if not text.strip():
            return result

        # Parse the document
        try:
            parser = self._get_parser()
            lexer = self._create_lexer()
            ast = parser.parser.parse(text, lexer=lexer.lexer)
            result.parse_successful = True
            result.ast = ast

            # Extract symbols from AST
            result.symbols = self._extract_symbols(ast, text)

            # Perform semantic analysis
            semantic_diagnostics = self._semantic_analysis(ast, text)
            result.diagnostics.extend(semantic_diagnostics)

        except CompileError as e:
            # Extract line number from error
            diagnostic = self._compile_error_to_diagnostic(e, text)
            result.diagnostics.append(diagnostic)

        except Exception as e:
            # Generic parse error
            diagnostic = self._exception_to_diagnostic(e, text)
            result.diagnostics.append(diagnostic)

        # Add syntax-based diagnostics (fast, regex-based)
        syntax_diagnostics = self._syntax_diagnostics(text)
        result.diagnostics.extend(syntax_diagnostics)

        return result

    def _compile_error_to_diagnostic(
        self,
        error: CompileError,
        text: str,
    ) -> Diagnostic:
        """Convert a CompileError to a Diagnostic."""
        # Try to extract line number
        line = 0
        col = 0

        if hasattr(error, "line_num") and error.line_num is not None:
            line = error.line_num - 1  # Convert to 0-indexed
        else:
            # Try to extract from message
            match = re.search(r"line (\d+)", str(error), re.IGNORECASE)
            if match:
                line = int(match.group(1)) - 1

        # Get line content for range calculation
        lines = text.split("\n")
        if 0 <= line < len(lines):
            col = 0
            end_col = len(lines[line])
        else:
            end_col = 0

        return Diagnostic(
            range=Range(
                start=Position(line=line, character=col),
                end=Position(line=line, character=end_col),
            ),
            message=str(error),
            severity=DiagnosticSeverity.ERROR,
            code="compile-error",
            source="nss",
        )

    def _exception_to_diagnostic(
        self,
        error: Exception,
        text: str,
    ) -> Diagnostic:
        """Convert a generic exception to a Diagnostic."""
        # Try to extract line number from traceback or message
        line = 0

        error_str = str(error)
        match = re.search(r"line (\d+)", error_str, re.IGNORECASE)
        if match:
            line = int(match.group(1)) - 1

        # Also check for "position" or "lineno" in the error
        pos_match = re.search(r"position (\d+)", error_str, re.IGNORECASE)
        if pos_match and line == 0:
            # Convert character position to line number
            pos = int(pos_match.group(1))
            current_pos = 0
            for i, content in enumerate(text.split("\n")):
                if current_pos + len(content) + 1 >= pos:
                    line = i
                    break
                current_pos += len(content) + 1

        lines = text.split("\n")
        if 0 <= line < len(lines):
            end_col = len(lines[line])
        else:
            end_col = 0

        return Diagnostic(
            range=Range(
                start=Position(line=line, character=0),
                end=Position(line=line, character=end_col),
            ),
            message=f"Parse error: {error}",
            severity=DiagnosticSeverity.ERROR,
            code="parse-error",
            source="nss",
        )

    def _syntax_diagnostics(self, text: str) -> list[Diagnostic]:
        """Quick syntax-based diagnostics (no parsing required)."""
        diagnostics: list[Diagnostic] = []
        lines = text.split("\n")

        # Track brace balance
        brace_balance = 0
        paren_balance = 0

        for line_num, line in enumerate(lines):
            stripped = line.strip()

            # Skip comments
            if stripped.startswith("//"):
                continue

            # Remove string literals for analysis
            line_no_strings = re.sub(r'"[^"\\]*(\\.[^"\\]*)*"', '""', stripped)

            # Track balance
            brace_balance += line_no_strings.count("{") - line_no_strings.count("}")
            paren_balance += line_no_strings.count("(") - line_no_strings.count(")")

            # Check for double semicolons
            if ";;" in line_no_strings:
                diagnostics.append(
                    Diagnostic(
                        range=Range(
                            start=Position(line=line_num, character=line.find(";;")),
                            end=Position(line=line_num, character=line.find(";;") + 2),
                        ),
                        message="Redundant semicolon",
                        severity=DiagnosticSeverity.WARNING,
                        code="redundant-semicolon",
                    )
                )

            # Check for common typos
            if re.search(r"\bvodi\b", stripped, re.IGNORECASE):
                match = re.search(r"\bvodi\b", stripped, re.IGNORECASE)
                if match:
                    diagnostics.append(
                        Diagnostic(
                            range=Range(
                                start=Position(line=line_num, character=match.start()),
                                end=Position(line=line_num, character=match.end()),
                            ),
                            message="Did you mean 'void'?",
                            severity=DiagnosticSeverity.ERROR,
                            code="unknown-type",
                            suggestions=["void"],
                        )
                    )

        # Check for unbalanced braces
        if brace_balance > 0:
            diagnostics.append(
                Diagnostic(
                    range=Range(
                        start=Position(line=len(lines) - 1, character=0),
                        end=Position(line=len(lines) - 1, character=0),
                    ),
                    message=f"Missing {brace_balance} closing brace(s) '}}'",
                    severity=DiagnosticSeverity.ERROR,
                    code="unbalanced-braces",
                )
            )
        elif brace_balance < 0:
            diagnostics.append(
                Diagnostic(
                    range=Range(
                        start=Position(line=0, character=0),
                        end=Position(line=0, character=0),
                    ),
                    message=f"Extra {-brace_balance} closing brace(s) '}}'",
                    severity=DiagnosticSeverity.ERROR,
                    code="unbalanced-braces",
                )
            )

        return diagnostics

    def _semantic_analysis(
        self,
        ast: CodeRoot,
        text: str,
    ) -> list[Diagnostic]:
        """Perform semantic analysis on the AST."""
        diagnostics: list[Diagnostic] = []

        # Check for entry point
        has_main = False
        has_starting_conditional = False

        for obj in ast.objects:
            if isinstance(obj, FunctionDefinition):
                if obj.identifier.identifier == "main":
                    has_main = True
                elif obj.identifier.identifier == "StartingConditional":
                    has_starting_conditional = True

        # Only warn about missing entry point if there are no forward declarations
        # (which would indicate this is an include file)
        has_forward_decl = any(isinstance(obj, FunctionForwardDeclaration) for obj in ast.objects)

        if not has_main and not has_starting_conditional and not has_forward_decl:
            diagnostics.append(
                Diagnostic(
                    range=Range(
                        start=Position(line=0, character=0),
                        end=Position(line=0, character=0),
                    ),
                    message="Script has no entry point (main() or StartingConditional()). This is normal for include files.",
                    severity=DiagnosticSeverity.INFORMATION,
                    code="no-entry-point",
                    suggestions=[
                        "Add void main() { } for executable scripts",
                        "Add int StartingConditional() { return TRUE; } for conditional scripts",
                    ],
                )
            )

        return diagnostics

    def _extract_symbols(
        self,
        ast: CodeRoot,
        text: str,
    ) -> list[DocumentSymbol]:
        """Extract document symbols from the AST."""
        symbols: list[DocumentSymbol] = []
        lines = text.split("\n")

        for obj in ast.objects:
            line_num = getattr(obj, "line_num", 1) - 1  # Convert to 0-indexed
            if line_num < 0:
                line_num = 0

            if isinstance(obj, FunctionDefinition):
                # Get function details
                return_type = str(getattr(obj.return_type, "builtin", "void")).split(".")[-1]
                params = getattr(obj, "parameters", [])
                param_str = ", ".join(
                    f"{getattr(p.datatype, 'builtin', '?').name if hasattr(getattr(p.datatype, 'builtin', '?'), 'name') else str(getattr(p.datatype, 'builtin', '?')).split('.')[-1]} {p.identifier}"
                    for p in params
                )

                # Calculate range
                end_line = line_num
                # Find the closing brace
                brace_count = 0
                for i in range(line_num, len(lines)):
                    brace_count += lines[i].count("{") - lines[i].count("}")
                    if brace_count == 0 and "{" in "".join(lines[line_num : i + 1]):
                        end_line = i
                        break

                # Create symbol with children for parameters
                func_symbol = DocumentSymbol(
                    name=obj.identifier.identifier,
                    kind="function",
                    range=Range(
                        start=Position(line=line_num, character=0),
                        end=Position(line=end_line, character=len(lines[end_line]) if end_line < len(lines) else 0),
                    ),
                    selection_range=Range(
                        start=Position(line=line_num, character=0),
                        end=Position(line=line_num, character=len(lines[line_num]) if line_num < len(lines) else 0),
                    ),
                    detail=f"{return_type} {obj.identifier.identifier}({param_str})",
                    children=[],
                )

                # Add parameters as children
                for param in params:
                    param_type = str(getattr(param.datatype, "builtin", "?"))
                    if "." in param_type:
                        param_type = param_type.split(".")[-1]
                    func_symbol.children.append(
                        DocumentSymbol(
                            name=param.identifier.identifier,
                            kind="parameter",
                            range=Range(
                                start=Position(line=line_num, character=0),
                                end=Position(line=line_num, character=0),
                            ),
                            selection_range=Range(
                                start=Position(line=line_num, character=0),
                                end=Position(line=line_num, character=0),
                            ),
                            detail=param_type,
                        )
                    )

                symbols.append(func_symbol)

            elif isinstance(obj, StructDefinition):
                members = getattr(obj, "members", [])

                struct_symbol = DocumentSymbol(
                    name=obj.identifier.identifier,
                    kind="struct",
                    range=Range(
                        start=Position(line=line_num, character=0),
                        end=Position(line=line_num, character=0),
                    ),
                    selection_range=Range(
                        start=Position(line=line_num, character=0),
                        end=Position(line=line_num, character=0),
                    ),
                    detail=f"struct ({len(members)} members)",
                    children=[],
                )

                # Add members
                for member in members:
                    member_type = str(getattr(member.datatype, "builtin", "?"))
                    if "." in member_type:
                        member_type = member_type.split(".")[-1]
                    struct_symbol.children.append(
                        DocumentSymbol(
                            name=member.identifier.identifier,
                            kind="variable",
                            range=Range(
                                start=Position(line=line_num, character=0),
                                end=Position(line=line_num, character=0),
                            ),
                            selection_range=Range(
                                start=Position(line=line_num, character=0),
                                end=Position(line=line_num, character=0),
                            ),
                            detail=member_type,
                        )
                    )

                symbols.append(struct_symbol)

            elif isinstance(obj, (GlobalVariableDeclaration, GlobalVariableInitialization)):
                var_type = str(getattr(obj.data_type, "builtin", "?"))
                if "." in var_type:
                    var_type = var_type.split(".")[-1]

                symbols.append(
                    DocumentSymbol(
                        name=obj.identifier.identifier,
                        kind="variable",
                        range=Range(
                            start=Position(line=line_num, character=0),
                            end=Position(line=line_num, character=0),
                        ),
                        selection_range=Range(
                            start=Position(line=line_num, character=0),
                            end=Position(line=line_num, character=0),
                        ),
                        detail=var_type,
                    )
                )

        return symbols

    def get_completions(
        self,
        text: str,
        line: int,
        character: int,
    ) -> list[CompletionItem]:
        """Get completion suggestions at a position.

        Args:
            text: Document text
            line: 0-indexed line number
            character: 0-indexed character position

        Returns:
            List of completion items
        """
        completions: list[CompletionItem] = []

        # Get the word being typed
        lines = text.split("\n")
        if line >= len(lines):
            return completions

        current_line = lines[line]
        # Find word start
        word_start = character
        while word_start > 0 and (current_line[word_start - 1].isalnum() or current_line[word_start - 1] == "_"):
            word_start -= 1

        prefix = current_line[word_start:character].lower() if character > word_start else ""

        # Add functions
        for func in self.functions:
            if not prefix or func.name.lower().startswith(prefix):
                return_type = getattr(func, "return_type", "void")
                params = getattr(func, "parameters", [])
                param_str = ", ".join(f"{getattr(p, 'type', '?')} {getattr(p, 'name', 'arg')}" for p in params[:3])
                if len(params) > 3:
                    param_str += ", ..."

                completions.append(
                    CompletionItem(
                        label=func.name,
                        kind="function",
                        detail=f"{return_type} {func.name}({param_str})",
                        documentation=getattr(func, "description", "") or "",
                        insert_text=f"{func.name}($0)",
                        sort_text=f"0_{func.name}",  # Functions first
                    )
                )

        # Add constants
        for const in self.constants:
            if not prefix or const.name.lower().startswith(prefix):
                const_type = getattr(const, "type", "")
                const_value = getattr(const, "value", "")

                completions.append(
                    CompletionItem(
                        label=const.name,
                        kind="constant",
                        detail=f"{const_type} = {const_value}" if const_value else const_type,
                        documentation=getattr(const, "description", "") or "",
                        insert_text=const.name,
                        sort_text=f"1_{const.name}",  # Constants second
                    )
                )

        # Add keywords
        keywords = [
            "void",
            "int",
            "float",
            "string",
            "object",
            "vector",
            "location",
            "effect",
            "event",
            "itemproperty",
            "talent",
            "action",
            "if",
            "else",
            "for",
            "while",
            "do",
            "switch",
            "case",
            "default",
            "break",
            "continue",
            "return",
            "struct",
            "const",
            "include",
            "TRUE",
            "FALSE",
            "OBJECT_SELF",
            "OBJECT_INVALID",
        ]
        for kw in keywords:
            if not prefix or kw.lower().startswith(prefix):
                completions.append(
                    CompletionItem(
                        label=kw,
                        kind="keyword",
                        detail="keyword",
                        insert_text=kw,
                        sort_text=f"2_{kw}",  # Keywords last
                    )
                )

        return completions

    def get_hover(
        self,
        text: str,
        line: int,
        character: int,
    ) -> HoverInfo | None:
        """Get hover information at a position.

        Args:
            text: Document text
            line: 0-indexed line number
            character: 0-indexed character position

        Returns:
            HoverInfo or None if nothing to show
        """
        lines = text.split("\n")
        if line >= len(lines):
            return None

        current_line = lines[line]

        # Find word at position
        word_start = character
        word_end = character

        while word_start > 0 and (current_line[word_start - 1].isalnum() or current_line[word_start - 1] == "_"):
            word_start -= 1
        while word_end < len(current_line) and (current_line[word_end].isalnum() or current_line[word_end] == "_"):
            word_end += 1

        word = current_line[word_start:word_end]
        if not word:
            return None

        word_lower = word.lower()

        # Search functions
        for func in self.functions:
            if func.name.lower() == word_lower:
                return_type = getattr(func, "return_type", "void")
                params = getattr(func, "parameters", [])
                param_str = ", ".join(f"{getattr(p, 'type', '?')} {getattr(p, 'name', 'arg')}" for p in params)
                description = getattr(func, "description", "") or ""

                content = f"```nwscript\n{return_type} {func.name}({param_str})\n```"
                if description:
                    content += f"\n\n{description}"

                return HoverInfo(
                    contents=content,
                    range=Range(
                        start=Position(line=line, character=word_start),
                        end=Position(line=line, character=word_end),
                    ),
                )

        # Search constants
        for const in self.constants:
            if const.name.lower() == word_lower:
                const_type = getattr(const, "type", "")
                const_value = getattr(const, "value", "")
                description = getattr(const, "description", "") or ""

                if const_type and const_value:
                    content = f"```nwscript\nconst {const_type} {const.name} = {const_value}\n```"
                elif const_value:
                    content = f"```nwscript\n{const.name} = {const_value}\n```"
                else:
                    content = f"```nwscript\n{const.name}\n```"

                if description:
                    content += f"\n\n{description}"

                return HoverInfo(
                    contents=content,
                    range=Range(
                        start=Position(line=line, character=word_start),
                        end=Position(line=line, character=word_end),
                    ),
                )

        return None

    def handle_request(self, request: dict[str, Any]) -> dict[str, Any]:
        """Handle a language server request.

        Request format:
        {
            'id': <unique request id>,
            'method': 'analyze' | 'completions' | 'hover' | 'shutdown',
            'params': {...}  # method-specific parameters
        }

        Response format:
        {
            'id': <same as request>,
            'result': {...} | None,
            'error': {'code': int, 'message': str} | None
        }
        """
        request_id = request.get("id")
        method = request.get("method", "")
        params = request.get("params", {})

        try:
            if method == "analyze":
                result = self.analyze(
                    text=params.get("text", ""),
                    filepath=params.get("filepath"),
                )
                return {
                    "id": request_id,
                    "result": self._analysis_result_to_dict(result),
                    "error": None,
                }

            elif method == "completions":
                completions = self.get_completions(
                    text=params.get("text", ""),
                    line=params.get("line", 0),
                    character=params.get("character", 0),
                )
                return {
                    "id": request_id,
                    "result": [self._completion_to_dict(c) for c in completions],
                    "error": None,
                }

            elif method == "hover":
                hover = self.get_hover(
                    text=params.get("text", ""),
                    line=params.get("line", 0),
                    character=params.get("character", 0),
                )
                return {
                    "id": request_id,
                    "result": self._hover_to_dict(hover) if hover else None,
                    "error": None,
                }

            elif method == "shutdown":
                return {
                    "id": request_id,
                    "result": {"status": "shutdown"},
                    "error": None,
                }

            elif method == "update_config":
                # Update functions/constants/library
                if "functions" in params:
                    self.functions = params["functions"]
                if "constants" in params:
                    self.constants = params["constants"]
                if "library" in params:
                    self.library = params["library"]
                if "is_tsl" in params:
                    self.is_tsl = params["is_tsl"]
                    # May need to reinitialize parser if game type changed
                    self._ensure_parser_initialized()

                return {
                    "id": request_id,
                    "result": {"status": "updated"},
                    "error": None,
                }

            else:
                return {
                    "id": request_id,
                    "result": None,
                    "error": {"code": -32601, "message": f"Unknown method: {method}"},
                }

        except Exception as e:
            return {
                "id": request_id,
                "result": None,
                "error": {"code": -32603, "message": str(e), "traceback": traceback.format_exc()},
            }

    def _analysis_result_to_dict(self, result: AnalysisResult) -> dict[str, Any]:
        """Convert AnalysisResult to dictionary for JSON serialization."""
        return {
            "diagnostics": [self._diagnostic_to_dict(d) for d in result.diagnostics],
            "symbols": [self._symbol_to_dict(s) for s in result.symbols],
            "parse_successful": result.parse_successful,
        }

    def _diagnostic_to_dict(self, diagnostic: Diagnostic) -> dict[str, Any]:
        """Convert Diagnostic to dictionary."""
        return {
            "range": {
                "start": {"line": diagnostic.range.start.line, "character": diagnostic.range.start.character},
                "end": {"line": diagnostic.range.end.line, "character": diagnostic.range.end.character},
            },
            "message": diagnostic.message,
            "severity": int(diagnostic.severity),
            "code": diagnostic.code,
            "source": diagnostic.source,
            "suggestions": diagnostic.suggestions,
        }

    def _symbol_to_dict(self, symbol: DocumentSymbol) -> dict[str, Any]:
        """Convert DocumentSymbol to dictionary."""
        return {
            "name": symbol.name,
            "kind": symbol.kind,
            "range": {
                "start": {"line": symbol.range.start.line, "character": symbol.range.start.character},
                "end": {"line": symbol.range.end.line, "character": symbol.range.end.character},
            },
            "selection_range": {
                "start": {"line": symbol.selection_range.start.line, "character": symbol.selection_range.start.character},
                "end": {"line": symbol.selection_range.end.line, "character": symbol.selection_range.end.character},
            },
            "detail": symbol.detail,
            "children": [self._symbol_to_dict(c) for c in symbol.children],
        }

    def _completion_to_dict(self, completion: CompletionItem) -> dict[str, Any]:
        """Convert CompletionItem to dictionary."""
        return {
            "label": completion.label,
            "kind": completion.kind,
            "detail": completion.detail,
            "documentation": completion.documentation,
            "insert_text": completion.insert_text,
            "sort_text": completion.sort_text,
        }

    def _hover_to_dict(self, hover: HoverInfo) -> dict[str, Any]:
        """Convert HoverInfo to dictionary."""
        result: dict[str, Any] = {"contents": hover.contents}
        if hover.range:
            result["range"] = {
                "start": {"line": hover.range.start.line, "character": hover.range.start.character},
                "end": {"line": hover.range.end.line, "character": hover.range.end.character},
            }
        return result

    @staticmethod
    def run_server(
        request_queue: Queue,
        response_queue: Queue,
        is_tsl: bool = False,
        functions: list[ScriptFunction] | None = None,
        constants: list[ScriptConstant] | None = None,
        library: dict[str, bytes] | None = None,
    ):
        """Run the language server in a subprocess.

        This is the entry point for the subprocess. It creates a language
        server instance and processes requests from the queue until shutdown.

        Args:
            request_queue: Queue to receive requests from
            response_queue: Queue to send responses to
            is_tsl: Whether this is for TSL (K2) or K1
            functions: Built-in script functions
            constants: Built-in script constants
            library: Script library for includes
        """
        # Load default functions/constants if not provided
        if functions is None or constants is None or library is None:
            # Import here to avoid circular imports and speed up initial import
            from pykotor.common.scriptdefs import (
                KOTOR_CONSTANTS,
                KOTOR_FUNCTIONS,
                TSL_CONSTANTS,
                TSL_FUNCTIONS,
            )
            from pykotor.common.scriptlib import KOTOR_LIBRARY, TSL_LIBRARY

            if is_tsl:
                functions = list(TSL_FUNCTIONS) if functions is None else functions
                constants = list(TSL_CONSTANTS) if constants is None else constants
                library = dict(TSL_LIBRARY) if library is None else library
            else:
                functions = list(KOTOR_FUNCTIONS) if functions is None else functions
                constants = list(KOTOR_CONSTANTS) if constants is None else constants
                library = dict(KOTOR_LIBRARY) if library is None else library

        # Create server instance (this initializes the parser - ~900ms one-time cost)
        server = NSSLanguageServer(
            functions=functions,
            constants=constants,
            library=library,
            is_tsl=is_tsl,
        )

        # Signal that server is ready
        response_queue.put(
            {
                "id": None,
                "result": {"status": "ready"},
                "error": None,
            }
        )

        # Process requests
        while True:
            try:
                request = request_queue.get(timeout=1.0)
            except Exception:
                # Timeout - check if parent process is still alive
                continue

            if request is None:
                # Shutdown signal
                break

            response = server.handle_request(request)
            response_queue.put(response)

            if request.get("method") == "shutdown":
                break
